import platform
import socket
import subprocess
import fileinput
import multiprocessing
import os
import re
import sys
import pwd

class mac_proxy:
    '''
    Refers to the macos version of system wide proxy. Refer to `Proxy` class for initializing system wide proxy.
    '''

    def __init__(self, ip_address, port):
        self.ip_address = ip_address
        self.port       = port

    def join(self):
        pass

class lin_proxy:
    '''
    Refers to the linux version of system wide proxy. Refer to `Proxy` class for initializing system wide proxy.
    '''

    def __init__(self, ip_address, port):
        self.aptconf    = '/etc/apt/apt.conf'
        self.envconf    = '/etc/environment'
        self.bashconf   = '/etc/bash.bashrc'
        self.wgetconf   = '/etc/wgetrc'
        self.ip_address = ip_address
        self.port       = str(port)

        out = subprocess.call("gsettings list-recursively org.gnome.system.proxy", stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        self.gnome = True if not out else False
        if os.geteuid() != 0:
            raise OSError("This library requires root privileges")

        if not os.path.isfile(self.aptconf):
            fl = open(self.aptconf, 'w')
            fl.close()

    def set_env_vars(self):
        exists = False
        for line in fileinput.input(self.envconf, inplace=1):
            if "proxy" in line:
                exists = True
                line = re.sub(r'(.*)_proxy=(.*)', r'\1_proxy="\1://'+self.ip_address+':'+self.port+"/\"\n", line.rstrip())
            sys.stdout.write(line)

        if not exists:
            fl = open(self.envconf, "w")
            fl.write("PATH=\"/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/usr/games:/usr/local/games\"\n")
            fl.write("http_proxy=\"http://"+self.ip_address+":"+self.port+"/\"\n")
            fl.write("https_proxy=\"https://"+self.ip_address+":"+self.port+"/\"\n")
            fl.write("ftp_proxy=\"ftp://"+self.ip_address+":"+self.port+"/\"\n")
            fl.write("socks_proxy=\"socks://"+self.ip_address+":"+self.port+"/\"\n")
            fl.close()

    def set_apt_vars(self):
        exists = False
        for line in fileinput.input(self.aptconf, inplace=1):
            if "Acquire::" in line and "Cache" not in line:
                exists = True
                line = re.sub(r'Acquire::(.*)::proxy (.*)', r'Acquire::\1::proxy "\1://'+self.ip_address+":"+self.port+"/\";\n", line.rstrip())
            sys.stdout.write(line)

        if not exists:
            fl = open(self.aptconf, "w")
            fl.write("Acquire::http::proxy \"http://"+self.ip_address+":"+self.port+"/\";\n")
            fl.write("Acquire::https::proxy \"https://"+self.ip_address+":"+self.port+"/\";\n")
            fl.write("Acquire::ftp::proxy \"ftp://"+self.ip_address+":"+self.port+"/\";\n")
            fl.write("Acquire::http::No-Cache \"True\";\n")
            fl.write("Acquire::socks::proxy \"socks://"+self.ip_address+":"+self.port+"/\";\n")
            fl.close()

    def set_bash_vars(self):
        exists = False
        for line in fileinput.input(self.bashconf, inplace=1):
            if "export" in line:
                exists = True
                line = re.sub(r'export (.*)_proxy=(.*)', r'export \1_proxy=\1://'+self.ip_address+':'+self.port+'\n', line.rstrip())
            sys.stdout.write(line)

        if not exists:
            fl = open(self.bashconf, "r+")
            l  = fl.readlines()
            fl.close()

            fl = open(self.bashconf, "a")

            if not l[-1][-1]=='\n':
                fl.write("\n")

            fl.write("export http_proxy=http://"+self.ip_address+":"+self.port+"\n")
            fl.write("export https_proxy=https://"+self.ip_address+":"+self.port+"\n")
            fl.write("export ftp_proxy=ftp://"+self.ip_address+":"+self.port+"\n")
            fl.close()

    def set_wget_vars(self):
        exists = False
        for line in fileinput.input(self.wgetconf, inplace=1):
            if not line.startswith("#") and "proxy" in line:
                exists = True
                line = re.sub(r'(.*)_proxy=(.*)//(.*)', r'\1_proxy=\1://'+self.ip_address+':'+self.port+'\n', line.rstrip())
            sys.stdout.write(line)

        if not exists:
            fl = open(self.wgetconf, "r+")
            l  = fl.readlines()
            fl.close()

            fl = open(self.wgetconf, "a")
            if not l[-1][-1]=='\n':
                fl.write("\n")

            fl.write("http_proxy=http://"+self.ip_address+":"+self.port+"\n")
            fl.write("https_proxy=https://"+self.ip_address+":"+self.port+"\n")
            fl.write("ftp_proxy=ftp://"+self.ip_address+":"+self.port+"\n")

    def set_gsettings(self):
        def caller(ip_address, port, uid, gid, dhome, logname):
            os.setgid(gid)
            os.setuid(uid)
            os.environ['HOME']    = dhome
            os.environ['LOGNAME'] = logname

            subprocess.call(f"dbus-launch gsettings set org.gnome.system.proxy mode 'manual'", shell=True)
            subprocess.call(f"dbus-launch gsettings set org.gnome.system.proxy.http host '{ip_address}'", shell=True)
            subprocess.call(f"dbus-launch gsettings set org.gnome.system.proxy.http port {port}", shell=True)
            subprocess.call(f"dbus-launch gsettings set org.gnome.system.proxy.https host '{ip_address}'", shell=True)
            subprocess.call(f"dbus-launch gsettings set org.gnome.system.proxy.https port {port}", shell=True)
            subprocess.call(f"dbus-launch gsettings set org.gnome.system.proxy use-same-proxy true", shell=True)

        users = pwd.getpwall()
        for user in users:
            if hasattr(user, 'pw_name') and hasattr(user, 'pw_shell'):
                if (user.pw_shell == "/bin/bash" or user.pw_shell == "bin/sh") and user.pw_uid != 0:
                    mp = multiprocessing.Process(target=caller, args=(self.ip_address, self.port, user.pw_uid, user.pw_gid, user.pw_dir, user.pw_name))
                    mp.daemon = True
                    mp.start()
                    mp.join()

    def rem_wget_vars(self):
        with open(self.wgetconf, "r+") as f:
            new_f = f.readlines()
            f.seek(0)
            for line in new_f:
                if "http_proxy" not in line and "https_proxy" not in line and "ftp_proxy" not in line and "socks_proxy" not in line:
                    f.write(line)
            f.truncate()
            f.close()

    def rem_bash_vars(self):
        with open(self.bashconf, "r+") as f:
            new_f = f.readlines()
            f.seek(0)
            for line in new_f:
                if "export http_proxy" not in line and "export https_proxy" not in line and "export ftp_proxy" not in line and "export socks_proxy" not in line:
                    f.write(line)
            f.truncate()
            f.close()

    def rem_env_vars(self):
        with open(self.envconf, "r+") as f:
            new_f = f.readlines()
            f.seek(0)
            for line in new_f:
                if "http_proxy" not in line and "https_proxy" not in line and "ftp_proxy" not in line and "socks_proxy" not in line:
                    f.write(line)
            f.truncate()
            f.close()

    def rem_apt_vars(self):
        with open(self.aptconf, "r+") as f:
            new_f = f.readlines()
            f.seek(0)
            for line in new_f:
                if "Acquire::http" not in line and "Acquire::https" not in line and "Acquire::ftp" not in line and "Acquire::socks" not in line:
                    f.write(line)
            f.truncate()
            f.close()

    def rem_gsettings(self):
        def caller(ip_address, port, uid, gid, dhome, logname):
            os.setgid(gid)
            os.setuid(uid)
            os.environ['HOME']    = dhome
            os.environ['LOGNAME'] = logname

            subprocess.call(f"dbus-launch gsettings set org.gnome.system.proxy mode 'none'", shell=True)

        users = pwd.getpwall()
        for user in users:
            if hasattr(user, 'pw_name') and hasattr(user, 'pw_shell'):
                if (user.pw_shell == "/bin/bash" or user.pw_shell == "bin/sh") and user.pw_uid != 0:
                    mp = multiprocessing.Process(target=caller, args=(self.ip_address, self.port, user.pw_uid, user.pw_gid, user.pw_dir, user.pw_name))
                    mp.daemon = True
                    mp.start()
                    mp.join()

    def join(self):
        if self.gnome:
            self.set_gsettings()
        else:
            self.set_env_vars()
            self.set_apt_vars()
            self.set_bash_vars()
            self.set_wget_vars()

    def del_proxy(self):
        if self.gnome:
            self.rem_gsettings()
        else:
            self.rem_wget_vars()
            self.rem_bash_vars()
            self.rem_apt_vars()
            self.rem_env_vars()

class win_proxy:
    '''
    Refers to the windows version of system wide proxy. Refer to `Proxy` class for initializing system wide proxy.
    '''

    def __init__(self, ip_address, port):
        self.glob_import('winreg')
        self.glob_import('ctypes')

        self.regkey = winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Software\Microsoft\Windows\CurrentVersion\Internet Settings', 0, winreg.KEY_ALL_ACCESS)
        self.internet_option_refresh = 37
        self.internet_option_settings_changed = 39
        self.internet_set_option = ctypes.windll.Wininet.InternetSetOptionW

        self.ip_address = ip_address
        self.port      = port

    def glob_import(self, module_name):
        globals()[module_name] = __import__(module_name)

    def refresh(self):
        self.internet_set_option(0, self.internet_option_settings_changed, 0, 0)
        self.internet_set_option(0, self.internet_option_refresh, 0, 0)

    def set_key(self, name, value):
        try:
            _, reg_type = winreg.QueryValueEx(self.regkey, name)
            winreg.SetValueEx(self.regkey, name, 0, reg_type, value)
        except FileNotFoundError:
            winreg.SetValueEx(self.regkey, name, 0, winreg.REG_SZ, value)

    def set_proxy(self):
        try:
            self.set_key('ProxyEnable', 1)
            #self.set_key('ProxyOverride', u'*.local;<local>')
            self.set_key('ProxyServer', u'%s:%i' % (self.ip_address, self.port))
            return True
        except IndexError:
            raise ValueError(f"Unable to find the registry path for proxy")
            return False

    def del_proxy(self):
        try:
            self.set_key('ProxyEnable', 0)
        except FileNotFoundError:
            pass

    def join(self):
        if not self.set_proxy():
            raise ValueError(f"Error setting proxy credentials")

        self.refresh()

class Proxy:

    ## Gets necessary information and kill mitmproxy if already running
    def __init__(self, ip_address, port):
        '''
            Accepts two arguments:

            ip_address: The IP address for system wide proxy
            port: The port for system wide proxy
        '''
        self.ip_address = ip_address
        self.port = port

    def get_prox_instance(self):
        plat = platform.system().lower()
        if plat == "windows":
            prox = win_proxy(self.ip_address, self.port)
        elif plat == "linux":
            prox = lin_proxy(self.ip_address, self.port)
        elif plat == "macos":
            prox = mac_proxy(self.ip_address, self.port)
        else:
            raise OSError("Unable to determine the underlying operating system")

        return prox

    def engage(self):
        '''
            Setup system wide proxy.
        '''

        prox = self.get_prox_instance()
        prox.join()

    def cleanup(self):
        '''
            Removes system wide proxy
        '''

        prox = self.get_prox_instance()
        prox.del_proxy()

        if hasattr(prox, 'refresh'):
            prox.refresh()



from proxy.http.parser import HttpParser, httpParserTypes, httpParserStates
from proxy.http.proxy import HttpProxyBasePlugin

class PluginBase(HttpProxyBasePlugin):
    '''Modified HttpProxyBasePlugin from proxy.py. Provides more functionality'''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request  = HttpParser(httpParserTypes.REQUEST_PARSER)
        self.response = HttpParser(httpParserTypes.RESPONSE_PARSER)

    def before_upstream_connection(self, connection):
        return self.intercept_connection(connection)

    def handle_client_request(self, request):
        rtval = self.intercept_request(request)
        self.request = rtval
        return rtval

    def handle_upstream_chunk(self, chunk):
        self.response.parse(chunk.tobytes())
        if self.response.state == httpParserStates.COMPLETE:
            self.intercept_response(self.response)
        return memoryview(self.intercept_chunk(chunk.tobytes()))

    def on_upstream_connection_close(self):
        self.close_connection()

    def intercept_connection(self, conn):
        '''
            Intercept Connection. The `conn` argument represents
            HttpParser class from proxy.http.parser.

            Intercept the initial connection state and represent
            the data before the connection is actuall established.

            For TLS Interception, it would most probably be a
            CONNECT request. You can check host, port and other initials
            in this section.

            Modified/Orignal `conn` must be returned in this function.
            Returning `None` will drop the connection.
        '''
        return conn

    def intercept_request(self, request):
        '''
            Intercept Request. The `request` argument represents
            HttpParser class from proxy.http.parser

            Intercept the orignal request sent to the server. The
            request can be modified at this stage.

            You will see GET/POST/HEAD and other method requests
            here. You can modify the request body and other
            necessary details here.

            Modified/Orignal `request` must be returned in this function.
            Returning `None` will not send the request to the server.
        '''
        return request

    def intercept_chunk(self, chunk):
        '''
            Intercept chunks received from the server. The `chunk`
            argument represents built-in `bytes` object from
            python.

            Intercept response in chunks received from server.
            It is not in the final form yet. So, you can the received
            response here

            You will see mostly plain bytes in this section with what's
            incoming. This chunk from this function is then appended
            to response in `interept_response`.

            Modified/Orignal `chunk` must be returned in this function.
            Empty bytes object will append nothing to response
            Return `None` will cancel the response at that very moment.
        '''
        return chunk

    def intercept_response(self, response):
        '''
            Intercept response received from the server. The `response`
            argument represents HttpParser from proxy.http.parser.

            Intercept the response combined from the chunks received in
            `intercept_chunk` method. The response can only be viewed
            at this stage. It is not to be modified here

            NOTE: Changing or returning response here won't have any effect
            on the orignal response. For modifying response see
            `intercept_response` method.

            This function doesn't return anything.
        '''
        return response

    def close_connection(self):
        '''
            This methods gets called when the connection has been closed.
            It doesn't have arguments.

            The request and response both can be analyzed at this part
            and can be accessed at `self.request` and `self.response`
            respectively.

            The modifications in here won't mean anything. But you can
            deduce other results in this section.

            This function doesn't return anything and basically used for
            analysis.
        '''
        pass



import subprocess
import os
import platform

class Installer:

    def __init__(self, output=os.devnull):
        self.chocoline = 'powershell.exe "Set-ExecutionPolicy Bypass -Scope Process -Force; [System.Net.ServicePointManager]::SecurityProtocol = [System.Net.ServicePointManager]::SecurityProtocol -bor 3072; iex ((New-Object System.Net.WebClient).DownloadString(\'https://community.chocolatey.org/install.ps1\'))'
        self.chocolime = 'powershell.exe "choco install {} -y"'
        self.platform  = platform.system().lower()
        if self.platform == "windows":
            self.startupinfo = subprocess.STARTUPINFO()
            self.startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        self.output    = output

    def activate(self):
        fl = open(self.output, 'w')
        if hasattr(self, 'startupinfo'):
            cmline = subprocess.Popen(
                self.chocoline,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=fl,
                stderr=fl,
                startupinfo=self.startupinfo
            )
        else:
            cmline = subprocess.Popen(
                self.chocoline,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=fl,
                stderr=fl
            )
        cmline.communicate()
        fl.close()

        if cmline.returncode:
            raise ImportError(f"Unable to install choco installer. Return code: {cmline}")

        return cmline.returncode == 0

    def install(self, pkg):
        fl = open(self.output, 'a')
        cmline = subprocess.Popen(
            self.chocolime.format(pkg),
            shell=True,
            stdin=subprocess.PIPE,
            stdout=fl,
            stderr=fl,
            startupinfo=self.startupinfo
        )
        cmline.communicate()
        fl.close()

        if cmline.returncode:
            raise ImportError(f"Unable to install {pkg}. Return code: {cmline}")

        return cmline.returncode == 0



from OpenSSL import crypto, SSL
import OpenSSL
import random
import re
import os
import shutil
import platform
import subprocess

class Generator:

    def __init__(self, email=None, country=None, province=None, locality=None, organization=None, unit=None, commonname=None):
        self.email   = email
        self.country = country
        self.province = province
        self.locality = locality
        self.organization = organization
        self.unit     = unit
        self.commonname = commonname or "example.com"
        self.serial_number = random.getrandbits(64)
        self.valid_start = 0
        self.valid_end   = 10*365*24*60*60

    def generate(self):
        self.cert = crypto.X509()
        self.key = crypto.PKey()
        self.key.generate_key(crypto.TYPE_RSA, 2048)
        if self.country:
            self.cert.get_subject().C = self.country
        if self.province:
            self.cert.get_subject().ST = self.province
        if self.locality:
            self.cert.get_subject().L = self.locality
        if self.organization:
            self.cert.get_subject().O = self.organization
        if self.unit:
            self.cert.get_subject().OU = self.unit
        self.cert.get_subject().CN = self.commonname
        if self.email:
            self.cert.get_subject().emailAddress = self.email
        self.cert.set_serial_number(self.serial_number)
        self.cert.gmtime_adj_notBefore(self.valid_start)
        self.cert.gmtime_adj_notAfter(self.valid_end)
        self.cert.set_issuer(self.cert.get_subject())
        self.cert.set_pubkey(self.key)
        self.cert.sign(self.key, 'sha256')

    def gen_key(self, key_file):
        key_file = open(key_file, 'wt')
        try:
            key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, self.key).decode("utf-8"))
        except TypeError:
            key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, self.key))
        key_file.close()

    def gen_cert(self, cert_file):
        cert_file = open(cert_file, 'wt')
        try:
            cert_file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, self.cert).decode("utf-8"))
        except TypeError:
            cert_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, self.key))
        cert_file.close()

    def gen_pfx(self, pfx_file):
        p12 = OpenSSL.crypto.PKCS12()
        p12.set_privatekey( self.key )
        p12.set_certificate( self.cert )

        pfx_file = open(pfx_file, 'wb')
        pfx_file.write(p12.export())
        pfx_file.close()

class Importer:

    def __init__(self, home_paths):
        self.home_paths = home_paths

        if not os.path.isfile(self.home_paths['privname']):
            raise FileNotFoundError("No private key file was found in home directory. It has either been modified or deleted. ")

        if not os.path.isfile(self.home_paths['certname']):
            raise FileNotFoundError("No cert file was found in home directory. It has either been modified or deleted. ")

        if not os.path.isfile(self.home_paths['pfxname']):
            raise FileNotFoundError("No pfx file was found in home directory. It has either been modified or deleted. ")

    def __config_wn_firefox(self):
        import winreg
        import ctypes

        is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0
        if not is_admin:
            raise PermissionError("Configuring firefox requires admin privileges")

        try:
            tkey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\firefox.exe")
            path = winreg.QueryValueEx(tkey, 'Path')[0]
        except:
            return -1

        autoconfig = os.path.join(path, 'defaults/pref/autoconfig.js')
        proxconfig = os.path.join(path, 'proxverter.cfg')

        fl = open(autoconfig, 'w')
        fl.write('pref("general.config.filename", "proxverter.cfg");\n')
        fl.write('pref("general.config.obscure_value", 0);\n')
        fl.close()

        fl = open(proxconfig, 'w')
        fl.write('//\n')
        fl.write('lockPref("network.proxy.type", 5);\n')
        fl.write('lockPref("security.enterprise_roots.enabled", true);\n')
        fl.close()

    def __import_windows(self):
        import ctypes

        is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0
        if not is_admin:
            raise PermissionError("Importing certificate requires admin privileges")

        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

        comm = subprocess.check_output(
            'powershell.exe "Get-PfxCertificate -FilePath {}"'.format(self.home_paths['pfxname']),
            shell=True,
            startupinfo=startupinfo,
            stdin=subprocess.PIPE
        )

        thumbprint = comm.split(b"\r\n")[3].split(b" ")[0].decode()

        comm = subprocess.check_output(
            'powershell.exe "Test-Path (Join-Path Cert:\LocalMachine\Root\ {})"'.format(thumbprint),
            shell=True,
            startupinfo=startupinfo,
            stdin=subprocess.PIPE
        )

        if comm.strip() == b"False":
            comm = subprocess.call(
                'powershell.exe "Import-PfxCertificate -FilePath \'{}\' -CertStoreLocation cert:\LocalMachine\Root\"'.format(self.home_paths['pfxname']),
                shell=True,
                startupinfo=startupinfo,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE
            )
            if comm:
                raise SystemError("Error while importing certificate ")

            return 0
        else:
            return 1

    def __import_linux(self):
        shutil.copy(self.home_paths['certname'], f"/usr/local/share/ca-certificates/proxverter.pem")
        shutil.copy(self.home_paths['certname'], f"/etc/ssl/certs/proxverter.pem")
        rtval = subprocess.call("update-ca-certificates", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return rtval

    def cimport(self):
        plat = platform.system().lower()
        if plat == "windows":
            rtval = self.__import_windows()
            self.__config_wn_firefox()
            return rtval
        elif plat == "linux":
            rtval = self.__import_linux()
            return rtval
        elif plat == "macos":
            pass
        else:
            raise OSError("Unable to determine the underlying operating system")



import setuptools

with open("README-pypi.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="proxverter",
    version="0.4.7",
    author="hash3liZer",
    author_email="admin@shellvoide.com",
    description="Cross platform system wide proxy server & TLS Interception library for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hash3liZer/Proxverter",
    project_urls={
        "Bug Tracker": "https://github.com/hash3liZer/Proxverter/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Operating System :: OS Independent",
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries',
        'Topic :: System :: Networking :: Firewalls',
        'Topic :: System :: Networking :: Monitoring',
    ],
    packages=['proxverter', 'proxverter.plugins'],
    python_requires=">=3.7",
    install_requires=open('requirements.txt', 'r').read().strip().split("\n"),
    keywords=[
        'http, https, proxy, transparent proxy, web proxy, https proxy',
        'python3 web proxy, python3 https'
    ]
)
